﻿using Infragistics.Controls.Editors;
using Infragistics.Documents.RichText;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;

namespace XRichTextEditor
{
    public class RichTextDocumentSerializer
    {
        private static string[] propertiesToSkip = new string[] { "IsSealed" };

        public static string Serialize(RichTextDocument document)
        {
            string result = string.Empty;

            foreach (var style in document.AvailableStyles)
            {
                result += PrintStyle(style);
            }

            result += PrintRootNode(document.RootNode);

            foreach (var node in document.RootNode.Body.ChildNodes)
            {
                result += ReadNode(node);
            }

            return result;
        }

        private static string PrintRootNode(DocumentRootNode rootNode)
        {
            string result = string.Empty;

            result += "RootNode:" + Environment.NewLine;

            result += PrintValueTypeProperties(rootNode, 1);

            if (rootNode.Settings != null)
            {
                result += CreateIndent(1) + "Settings:" + Environment.NewLine;

                result += PrintValueTypeProperties(rootNode.Settings, 2);

                result += PrintSubProperties(rootNode.Settings, 2, "DefaultCharacterSettings", "DefaultParagraphSettings", "ThemeFontLanguageSettings");
            }

            return result;
        }

        private static string PrintStyle(RichTextStyleBase style)
        {
            return PrintValueTypeProperties(style, 1).Insert(0, style.GetType().Name + Environment.NewLine);
        }

        public static void SaveToFile(RichTextDocument document, string filePath)
        {
            System.IO.File.WriteAllText(filePath, Serialize(document));
        }

        private static string ReadNode(NodeBase node)
        {
            string result = node.GetType().Name.ToString() + Environment.NewLine;

            if (node is RunNode)
            {
                result += PrintRunNode((RunNode)node);
            }
            else if (node is ParagraphNode)
            {
                result += PrintParagraphNode((ParagraphNode)node);
            }
            else if (node is TextNode)
            {
                result += PrintTextNodeProperties((TextNode)node);
            }
            else if (node is HyperlinkNode)
            {
                result += PrintHyperlinkNodeProperties((HyperlinkNode)node);
            }
            else if (node is TableNode)
            {
                result += PrintTableNode((TableNode)node);
            }
            else if (node is TableRowNode)
            {
                result += PrintTableRowNode((TableRowNode)node);
            }
            else if (node is TableCellNode)
            {
                result += PrintTableCellNode((TableCellNode)node);
            }

            if (node.HasChildNodes)
            {
                foreach (var n in node.ChildNodes)
                {
                    result += ReadNode(n);
                }
            }

            return result;
        }

        private static string PrintTableNode(TableNode node)
        {
            string result = string.Empty;

            result += PrintValueTypeProperties(node, 1);

            result += PrintSubProperties(node, 2, "Grid", "CustomAttributes");

            result += PrintSettingsString(node);

            if (node.Settings != null)
            {
                result += PrintValueTypeProperties(node.Settings);

                result += PrintSubProperties(node.Settings, 2, "FloatingSettings", "Borders", "CellMargins");
            }
            return result;
        }

        private static string PrintTableRowNode(TableRowNode node)
        {
            string result = string.Empty;

            result += PrintValueTypeProperties(node, 1);

            result += PrintSubProperties(node, 2, "SettingsResolved", "TableSettingsExceptions", "TableSettingsResolved", "CustomAttributes");

            result += PrintSettingsString(node);

            if (node.Settings != null)
                result += PrintValueTypeProperties(node.Settings);

            return result;
        }

        private static string PrintTableCellNode(TableCellNode node)
        {
            string result = string.Empty;

            result += PrintValueTypeProperties(node, 1);

            result += PrintSubProperties(node, 2, "SettingsResolved", "CustomAttributes");

            result += PrintSettingsString(node);

            if (node.Settings != null)
            {
                result += PrintValueTypeProperties(node.Settings);

                result += PrintSubProperties(node.Settings, 2, "Borders", "Margins");
            }
            return result;
        }

        private static string PrintRunNode(RunNode node)
        {
            string result = string.Empty;

            result += PrintValueTypeProperties(node, 1);

            result += PrintSubProperties(node, 2, "SettingsResolved", "CustomAttributes");

            result += PrintSettingsString(node);

            if (node.Settings != null)
            {
                result += PrintValueTypeProperties(node.Settings);

                result += PrintSubProperties(node.Settings, 2, "Border", "Color", "FontSettings", "LanguageSettings", "UnderlineColor");
            }
            return result;
        }

        private static string PrintParagraphNode(ParagraphNode node)
        {
            string result = string.Empty;

            result += PrintValueTypeProperties(node, 1);

            result += PrintSubProperties(node, 2, "MarkSettings", "ParagraphMarkSettingsResolved", "RunSettingsResolved", "SettingsResolved", "CustomAttributes", "ListItemInfoResolved");

            if (node.ListItemInfoResolved != null)
            {
                result += PrintSubProperties(node.ListItemInfoResolved, 3, "IndicatorSettingsResolved", "LevelDefinition", "Owner");
            }

            result += PrintSettingsString(node);

            if (node.Settings != null)
            {
                result += PrintValueTypeProperties(node.Settings);

                //TODO:Read TabStops
                result += PrintSubProperties(node.Settings, 2, "Borders", "Indentation", "Frame", "Spacing");
            }
            return result;
        }

        private static string PrintTextNodeProperties(TextNode node)
        {
            string result = string.Empty;

            result += PrintValueTypeProperties(node);

            result += PrintSubProperties(node, 1, "CustomAttributes");

            return result;
        }

        private static string PrintHyperlinkNodeProperties(HyperlinkNode node)
        {
            string result = string.Empty;

            result += PrintValueTypeProperties(node);

            result += PrintSubProperties(node, 1, "CustomAttributes");

            return result;
        }

        private static string PrintSubProperties(object obj, int startIndentCount, params string[] properties)
        {
            string result = string.Empty;

            foreach (var prop in properties)
            {
                var property = obj.GetType().GetProperty(prop);

                if (properties.Contains(property.Name))
                {
                    var propertyValue = property.GetValue(obj, null);

                    if (propertyValue != null)
                    {
                        result += CreateIndent(startIndentCount) + property.Name + ":" + Environment.NewLine;
                        result += PrintValueTypeProperties(propertyValue, startIndentCount + 1);
                    }
                }
            }

            return result;
        }

        private static string PrintValueTypeProperties(object node, int indentCount = 2)
        {
            string result = string.Empty;

            var props = node.GetType().GetProperties().Where(p => p.PropertyType.IsValueType || p.PropertyType == typeof(string) || p.PropertyType.IsEnum || p.PropertyType == typeof(object));

            //var allProps = node.GetType().GetProperties();
            //var notCovered = allProps.Except(props).ToList();

            foreach (var property in props)
            {
                if (propertiesToSkip.Contains(property.Name))
                    continue;
                var propertyValue = property.GetValue(node, null);

                if (propertyValue != null)
                {
                    result += CreateIndent(indentCount) + property.Name + ": " + propertyValue + Environment.NewLine;
                }
            }

            return result;
        }

        private static string PrintSettingsString(object node, int indentCount = 1)
        {
            string result = string.Empty;

            var settingsPropInfo = node.GetType().GetProperty("Settings");

            var settigsName = settingsPropInfo.PropertyType.Name;

            var settings = settingsPropInfo.GetValue(node, null);

            if (settings != null)
            {
                result = CreateIndent(indentCount) + settigsName + ":" + Environment.NewLine;
            }
            else
            {
                result = CreateIndent(indentCount) + settigsName + ": [null]" + Environment.NewLine;
            }

            return result;
        }

        private static string CreateIndent(int count)
        {
            string indent = "";

            for (int i = 0; i < count; i++)
            {
                indent += "\t";
            }

            return indent;
        }
    }
}